@@ -36,6 +36,8 @@ let init_flags () =
   begin
 (*** The following flags control search. ***)
     Hashtbl.add bool_flags "ENABLE_PATTERN_CLAUSES" false;
+    Hashtbl.add bool_flags "DYNAMIC_PATTERN_CLAUSES" true; (*** Create pattern clauses from universally quantified formulas during search. - Chad, April 6, 2011 ***)
+    Hashtbl.add bool_flags "PATTERN_CLAUSES_TRANSITIVITY_EQ" false; (*** Make pattern clauses for transitivity of equality. ***)
     Hashtbl.add bool_flags "PATTERN_CLAUSES_FORALL_AS_LIT" true;
     Hashtbl.add int_flags "PATTERN_CLAUSES_DELAY" 1;
     Hashtbl.add int_flags "PATTERN_CLAUSES_EQN_DELAY" 1;
@@ -112,6 +114,32 @@ let init_flags () =
     Hashtbl.add int_flags "POST_CONFRONT4_DELAY" 0;
     Hashtbl.add bool_flags "INITIAL_SUBTERMS_AS_INSTANTIATIONS" false;
     Hashtbl.add int_flags "PRIORITY_QUEUE_IMPL" 0; (*** Which version of priority queue implementation to use ***)
+    Hashtbl.add bool_flags "PREPROCESS_BEFORE_SPLIT" false;
+    Hashtbl.add bool_flags "TREAT_CONJECTURE_AS_SPECIAL" false;
+    Hashtbl.add int_flags "AXIOM_DELAY" 0; (*** Set this to > 0 to work on the negated conjecture first ***)
+    Hashtbl.add int_flags "RELEVANCE_DELAY" 0; (*** Set this to > 0 to delay axioms that do not share names with the negated conjecture. Only has an effect if TREAT_CONJECTURE_SPECIAL is true ***)
+    Hashtbl.add bool_flags "ALL_DEFS_AS_EQNS" false;
+    Hashtbl.add bool_flags "EAGERLY_PROCESS_INSTANTIATIONS" true;
+    Hashtbl.add int_flags "INSTANTIATION_DELAY" 1; (*** If EAGERLY_PROCESS_INSTANTIATIONS is false ***)
+    Hashtbl.add int_flags "ARTP_WEIGHT" 1; (*** If EAGERLY_PROCESS_INSTANTIATIONS is false ***)
+    Hashtbl.add int_flags "BASETP_WEIGHT" 1; (*** If EAGERLY_PROCESS_INSTANTIATIONS is false ***)
+    Hashtbl.add int_flags "OTP_WEIGHT" 1; (*** If EAGERLY_PROCESS_INSTANTIATIONS is false ***)
+    Hashtbl.add int_flags "AP_WEIGHT" 1; (*** If EAGERLY_PROCESS_INSTANTIATIONS is false ***)
+    Hashtbl.add int_flags "LAM_WEIGHT" 1; (*** If EAGERLY_PROCESS_INSTANTIATIONS is false ***)
+    Hashtbl.add int_flags "LAM_TP_WEIGHT_FAC" 1; (*** If EAGERLY_PROCESS_INSTANTIATIONS is false ***)
+    Hashtbl.add int_flags "NAME_WEIGHT" 1; (*** If EAGERLY_PROCESS_INSTANTIATIONS is false ***)
+    Hashtbl.add int_flags "NAME_TP_WEIGHT_FAC" 1; (*** If EAGERLY_PROCESS_INSTANTIATIONS is false ***)
+    Hashtbl.add int_flags "DB_WEIGHT" 1; (*** If EAGERLY_PROCESS_INSTANTIATIONS is false ***)
+    Hashtbl.add int_flags "DB_TP_WEIGHT_FAC" 1; (*** If EAGERLY_PROCESS_INSTANTIATIONS is false ***)
+    Hashtbl.add int_flags "FALSE_WEIGHT" 1; (*** If EAGERLY_PROCESS_INSTANTIATIONS is false ***)
+    Hashtbl.add int_flags "IMP_WEIGHT" 1; (*** If EAGERLY_PROCESS_INSTANTIATIONS is false ***)
+    Hashtbl.add int_flags "FORALL_WEIGHT" 1; (*** If EAGERLY_PROCESS_INSTANTIATIONS is false ***)
+    Hashtbl.add int_flags "FORALL_TP_WEIGHT_FAC" 1; (*** If EAGERLY_PROCESS_INSTANTIATIONS is false ***)
+    Hashtbl.add int_flags "EQ_WEIGHT" 1; (*** If EAGERLY_PROCESS_INSTANTIATIONS is false ***)
+    Hashtbl.add int_flags "EQ_TP_WEIGHT_FAC" 1; (*** If EAGERLY_PROCESS_INSTANTIATIONS is false ***)
+    Hashtbl.add int_flags "CHOICE_WEIGHT" 1; (*** If EAGERLY_PROCESS_INSTANTIATIONS is false ***)
+    Hashtbl.add int_flags "CHOICE_TP_WEIGHT_FAC" 1; (*** If EAGERLY_PROCESS_INSTANTIATIONS is false ***)
+
 (*** The following flags control translation to proof terms. ***)
     Hashtbl.add bool_flags "PFTRM_ADD_SYM_CLAUSES" true;
     Hashtbl.add bool_flags "PFTRM_PRESORT_CLAUSES" true;
