@@ -78,9 +78,11 @@ let print_subproof_info minclauses cri =
 	      match (Hashtbl.find cri c) with
 	      | InstRule(a,m,n) ->
 		  begin
+(***
 		    match a with
 		    | Ar(_,_) -> Printf.printf "HO Instantiation of type %s for\n%s:\n* %s\n" (stp_str a) (trm_str m) (trm_str n)
 		    | _ -> () (*** Printf.printf "FO Instantiation of type %s for\n%s:\n* %s\n" (stp_str a) (trm_str m) (trm_str n) ***)
+***)
 		  end
 	      | _ -> ()
 	    with
@@ -90,12 +92,12 @@ let print_subproof_info minclauses cri =
   end
     
 let rec new_assumption_clause c =
-  if (!verbosity > 8) then begin Printf.printf "Adding lits to minisat\n"; flush stdout end;
-  List.iter (fun l -> minisat_addLit l) c;
-  insert_assumption_clause c;
-  if (!verbosity > 8) then begin Printf.printf "Adding clause to minisat\n"; flush stdout end;
   try
     begin
+      insert_assumption_clause c;
+      if (!verbosity > 8) then begin Printf.printf "Adding lits to minisat\n"; flush stdout end;
+      List.iter (fun l -> minisat_addLit l) c;
+      if (!verbosity > 8) then begin Printf.printf "Adding clause to minisat\n"; flush stdout end;
       if (!verbosity > 4) then begin Printf.printf "About to call minisat_addClause (%s)\n" (List.fold_left (fun x y -> if ((String.length x) == 0) then (string_of_int y) else (x ^ " " ^ (string_of_int y))) "" c); flush stdout end;
       if (minisat_addClause()) then
 	begin
@@ -108,6 +110,7 @@ let rec new_assumption_clause c =
       else raise (Unsatisfiable None)
     end
   with
+  | DuplicateClause -> ()
   | (Unsatisfiable _) ->
       if (!verbosity > 20) then (Printf.printf "Proof found with assumption clauses only (no search)...%d clauses\n" (List.length (!clauses)); flush stdout);
       if (!mkproofterm) then
@@ -140,12 +143,12 @@ let filter lit =
 let filterneg lit = filter (- lit)
 
 let rec new_search_clause c r =
-  if (!verbosity > 8) then begin Printf.printf "Adding lits to minisat\n"; flush stdout end;
-  List.iter (fun l -> minisat_addLit l) c;
-  insert_search_clause c r;
-  if (!verbosity > 8) then begin Printf.printf "Adding clause to minisat\n"; flush stdout end;
   try
     begin
+      insert_search_clause c r;
+      if (!verbosity > 8) then begin Printf.printf "Adding lits to minisat\n"; flush stdout end;
+      List.iter (fun l -> minisat_addLit l) c;
+      if (!verbosity > 8) then begin Printf.printf "Adding clause to minisat\n"; flush stdout end;
       if (!verbosity > 4) then begin Printf.printf "About to call minisat_addClause (%s)\n" (List.fold_left (fun x y -> if ((String.length x) == 0) then (string_of_int y) else (x ^ " " ^ (string_of_int y))) "" c); flush stdout end;
       if (minisat_addClause()) then
 	begin
@@ -158,11 +161,14 @@ let rec new_search_clause c r =
       else raise (Unsatisfiable None)
     end
   with
+  | DuplicateClause -> ()
   | (Unsatisfiable _) ->
+      if (!verbosity > 4) then (Printf.printf "Proof found for a goal...%d clauses\n" (List.length (!clauses)); flush stdout);
       if (!mkproofterm) then
 	let cri = !clause_ruleinfo in (*** remember the clauses and this hashtable for them ***)
 (*** To test with minimal unsatisfiable core: ***)
 	let minclauses = minimal_unsatisfiable_core (!clauses) in
+	if (!verbosity > 4) then (Printf.printf "Reduced to %d clauses\n" (List.length minclauses); flush stdout);
 (*** To test all clauses: ***)
 (***	    let minclauses = !clauses in ***)
 	print_subproof_info minclauses cri;
@@ -190,21 +196,31 @@ let rec new_search_clause c r =
                  end)
               minclauses;
             List.iter (fun x -> Hashtbl.remove searchlits x) (!assumptionlits);
-            Printf.printf "Refutation Statistics:\n";
-            Printf.printf "Number of assumptions %d\n" (!numassumptionclauses);
-            Printf.printf "Number of search clauses %d\n" (!numsearchclauses);
-            Printf.printf "Number of search literals %d\n" (Hashtbl.length searchlits);
+	    if (!verbosity > 3) then
+	      begin
+		Printf.printf "Refutation Statistics:\n";
+		Printf.printf "Number of assumptions %d\n" (!numassumptionclauses);
+		Printf.printf "Number of search clauses %d\n" (!numsearchclauses);
+		Printf.printf "Number of search literals %d\n" (Hashtbl.length searchlits);
+	      end;
           end;
 	raise (Unsatisfiable (Some (SearchR (minclauses,Hashtbl.find cri))))
       else
 	raise (Unsatisfiable None)
 
-let pattern_clauses : (int * (metatrm option ref list) * (metatrm option ref list) * (metatrm list) * (int list) * dpair list * (metatrm list)) list ref
+let pattern_clauses : (int * (metatrm option ref list) * (metatrm option ref list) * (metatrm list) * (int list) * dpair list * (metatrm list option * bool)) list ref
     = ref []
 let pattern_clauses_usable : (int * trm) list ref
     = ref []
 exception TrivialClause
 
+(*** April 6, 2011 - Chad, I want to dynamically turn these on and off.  The immediate application is to make pattern clauses for transitivity of equation. ***)
+let pattern_clauses_forall_as_lit : bool ref = ref false 
+let pattern_clauses_onlyallstrict : bool ref = ref false 
+let pattern_clauses_eqnlits : bool ref = ref false 
+
+let pattern_clauses_transitivity_types : (stp,unit) Hashtbl.t = Hashtbl.create 10
+
 (*** The pattern clause has become ground.  Put it in the set of propositional clauses. ***)
 let apply_pattern_clause_2 nalllit evars abslits =
   if (!verbosity > 40) then
@@ -251,7 +267,7 @@ let allstrictp evars p =
     We replace the unmatched literal n1 =(a) n2 with C[n1] | -C[n2], partially instantiated.
     Since C[n1] is C[t], we add mlit to abslits.  We remove the unmatched equation n and replace it with -C[n2].
  ***)
-let rec apply_pattern_clause (nalllit,evars,locevars,unmatched,abslits,dlist,strictlits) mlit m =
+let rec apply_pattern_clause (nalllit,evars,locevars,unmatched,abslits,dlist,((strictlitsopt,rew) as misc)) mlit m =
   begin
     if (!verbosity > 40) then
       begin
@@ -261,41 +277,50 @@ let rec apply_pattern_clause (nalllit,evars,locevars,unmatched,abslits,dlist,str
 	  (String.concat " " (List.map string_of_int abslits));
 	Printf.printf "unmatched\n";
 	List.iter (fun m -> Printf.printf ". %s\n" (metatrm_str m)) unmatched;
-	Printf.printf "strict lits\n";
-	List.iter (fun m -> Printf.printf ". %s\n" (metatrm_str m)) strictlits
+	match strictlitsopt with
+	| Some(strictlits) ->
+	    begin
+	      Printf.printf "strict lits\n";
+	      List.iter (fun m -> Printf.printf ". %s\n" (metatrm_str m)) strictlits
+	    end
+	| _ -> ()
       end;
     if (not (List.mem (- mlit) abslits)) then (*** Unless it would be a trivial clause with both mlit and -mlit ***)
       begin
 	List.iter
 	  (fun n ->
-	    apply_pattern_clause_1 (nalllit,evars,locevars,unmatched,abslits,dlist,strictlits) n mlit m;
-	    if (get_bool_flag "PATTERN_CLAUSES_EQNLITS") then
+	    apply_pattern_clause_1 (nalllit,evars,locevars,unmatched,abslits,dlist,misc) n mlit m;
+	    if rew then
 	      begin
 		match n with
 		| MAp(MAp(MGround(Eq(a)),n1),n2) ->
-		    if (get_bool_flag "PATTERN_CLAUSES_ONLYALLSTRICT") then (*** Only match strict lits if this flag is set ***)
-		      begin
-			if (allstrictp evars n1) then (*** Note: If this flag is true, then evars and locevars are the same ***)
-			  apply_pattern_clause_eqn (nalllit,evars,locevars,unmatched,abslits,dlist,strictlits) n a n1 n2 mlit m (fun m -> m);
-			if (allstrictp evars n2) then
-			  apply_pattern_clause_eqn (nalllit,evars,locevars,unmatched,abslits,dlist,strictlits) n a n2 n1 mlit m (fun m -> m)
-		      end
-		    else
-		      begin
-			apply_pattern_clause_eqn (nalllit,evars,locevars,unmatched,abslits,dlist,strictlits) n a n1 n2 mlit m (fun m -> m);
-			apply_pattern_clause_eqn (nalllit,evars,locevars,unmatched,abslits,dlist,strictlits) n a n2 n1 mlit m (fun m -> m)
-		      end
+		    begin
+		      match strictlitsopt with
+		      | Some(strictlits)->
+			  begin
+			    if (allstrictp evars n1) then (*** Note: If this flag is true, then evars and locevars are the same ***)
+			      apply_pattern_clause_eqn (nalllit,evars,locevars,unmatched,abslits,dlist,misc) n a n1 n2 mlit m (fun m -> m);
+			    if (allstrictp evars n2) then
+			      apply_pattern_clause_eqn (nalllit,evars,locevars,unmatched,abslits,dlist,misc) n a n2 n1 mlit m (fun m -> m)
+			  end
+		      | _ ->
+			  begin
+			    apply_pattern_clause_eqn (nalllit,evars,locevars,unmatched,abslits,dlist,misc) n a n1 n2 mlit m (fun m -> m);
+			    apply_pattern_clause_eqn (nalllit,evars,locevars,unmatched,abslits,dlist,misc) n a n2 n1 mlit m (fun m -> m)
+			  end
+		    end
 		| _ -> ()
 	      end
 	  )
-	  (if (get_bool_flag "PATTERN_CLAUSES_ONLYALLSTRICT") then (*** Only match strict lits if this flag is set ***)
-	    strictlits
-	  else
-	    unmatched)
+	  begin
+	    match strictlitsopt with
+	    | Some strictlits -> strictlits
+	    | None -> unmatched
+	  end
       end
   end
 (*** n is an unmatched literal, try matching against m ***)
-and apply_pattern_clause_1 (nalllit,evars,locevars,unmatched,abslits,dlist,strictlits) n mlit m =
+and apply_pattern_clause_1 (nalllit,evars,locevars,unmatched,abslits,dlist,misc) n mlit m =
   try
     begin
       if (!verbosity > 40) then
@@ -374,8 +399,8 @@ and apply_pattern_clause_1 (nalllit,evars,locevars,unmatched,abslits,dlist,stric
 	unmatched;
       match (!unmatched2) with
       | (_::_) ->
-	  (*** If we are here, then the flag PATTERN_CLAUSES_ONLYALLSTRICT must be false and we need not keep up with strictlits.  Leave it empty here. ***)
-	  new_pattern_clause_3 (nalllit,List.rev (!evars2),!locevars2,!unmatched2,!abslits2,dlist2,[])
+	  (*** If we are here, then misc is (None,rew); can safely leave it alone. ***)
+	  new_pattern_clause_3 (nalllit,List.rev (!evars2),!locevars2,!unmatched2,!abslits2,dlist2,misc)
       | [] ->
 	  begin (*** clause has been fully instantiated ***)
 	    match (dlist2) with
@@ -398,7 +423,7 @@ and apply_pattern_clause_1 (nalllit,evars,locevars,unmatched,abslits,dlist,stric
   | TrivialClause -> ()
   | PatternMatchFailed -> ()
 (*** Rewrite with in m equation n:n1 =(a) n2 ***)
-and apply_pattern_clause_eqn ((nalllit,evars,locevars,unmatched,abslits,dlist,strictlits) as c) n a n1 n2 mlit m ctx =
+and apply_pattern_clause_eqn ((nalllit,evars,locevars,unmatched,abslits,dlist,misc) as c) n a n1 n2 mlit m ctx =
   let b = tpof m in
   if ((a = b) && (termNoDBs m)) then
     begin
@@ -479,8 +504,8 @@ and apply_pattern_clause_eqn ((nalllit,evars,locevars,unmatched,abslits,dlist,st
    ((metanorm (MAp(MAp(MGround(Imp),(ctx n2)),MGround(False))))::unmatched); (** Replace n with ~C[n2] **)
 	match (!unmatched2) with
 	| (_::_) ->
-	    (*** If we are here, then the flag PATTERN_CLAUSES_ONLYALLSTRICT must be false and we need not keep up with strictlits.  Leave it empty here. ***)
-	    new_pattern_clause_3 (nalllit,List.rev (!evars2),!locevars2,!unmatched2,!abslits2,dlist2,[])
+	  (*** If we are here, then misc is (None,rew); can safely leave it alone. ***)
+	    new_pattern_clause_3 (nalllit,List.rev (!evars2),!locevars2,!unmatched2,!abslits2,dlist2,misc)
 	| [] ->
 	    begin (*** clause has been fully instantiated ***)
 	      match (dlist2) with
@@ -513,7 +538,7 @@ and new_pattern_clause_3 c = (*** Process a new clause by trying it with every g
     begin
       Printf.printf "* new_pattern_clause_3 ";
       match c with
-      | (nalllit,evars,locevars,unmatched,abslits,dlist,strictlits) ->
+      | (nalllit,evars,locevars,unmatched,abslits,dlist,misc) ->
 	  begin
 	    Printf.printf "%d\n" nalllit;
 	    Printf.printf "evars\n";
@@ -546,8 +571,15 @@ and new_pattern_clause_3 c = (*** Process a new clause by trying it with every g
 		Printf.printf ". %s\n =? %s\n : %s\n" (metatrm_str m0) (trm_str n0) (stp_str b0)
 	      )
 	      dlist;
-	    Printf.printf "strict lits\n";
-	    List.iter (fun m -> Printf.printf ". %s\n" (metatrm_str m)) strictlits
+	    begin
+	      match misc with
+	      |	(Some strictlits,_) ->
+		  begin
+		    Printf.printf "strict lits\n";
+		    List.iter (fun m -> Printf.printf ". %s\n" (metatrm_str m)) strictlits
+		  end
+	      |	_ -> ()
+	    end
 	  end
     end;  
   List.iter (fun (mlit,m) -> apply_pattern_clause c mlit m)
@@ -556,7 +588,7 @@ and new_pattern_clause_3 c = (*** Process a new clause by trying it with every g
 
 (*** Create a new pattern clause, possibly computing the strict literals and eqn lits depending on flag settings. ***)
 let new_pattern_clause_3a (nalllit,evars,unmatched,abslits,dlist) =
-  if (get_bool_flag "PATTERN_CLAUSES_ONLYALLSTRICT") then
+  if (!pattern_clauses_onlyallstrict) then
     begin
       let strictlits = ref [] in
       List.iter
@@ -567,11 +599,11 @@ let new_pattern_clause_3a (nalllit,evars,unmatched,abslits,dlist) =
 	unmatched;
       match (!strictlits) with
       | (_::_) ->
-	  new_pattern_clause_3 (nalllit,evars,evars,unmatched,abslits,dlist,!strictlits)	  
+	  new_pattern_clause_3 (nalllit,evars,evars,unmatched,abslits,dlist,(Some !strictlits,!pattern_clauses_eqnlits))
       | [] -> () (*** No strict literals, so drop the pattern clause. ***)
     end
   else
-    new_pattern_clause_3 (nalllit,evars,evars,unmatched,abslits,dlist,[])
+    new_pattern_clause_3 (nalllit,evars,evars,unmatched,abslits,dlist,(None,!pattern_clauses_eqnlits))
 
 (*** Apply all pattern clauses so far against the negation of the new ground formula m ***)
 let apply_pattern_clauses mlit m =
@@ -629,7 +661,7 @@ let rec new_pattern_clauses_2 nalllit ml evars strict unmatched abslits =
 	begin (*** Treat formula as a literal in the clause (unless it's a forall, then whether or not we do depends on a flag.) ***)
 	  match m with
 	  | MAp(MGround(Forall(a)),m1) ->
-	      if (get_bool_flag "PATTERN_CLAUSES_FORALL_AS_LIT") then
+	      if (!pattern_clauses_forall_as_lit) then (*** Use pattern_clauses_forall_as_lit which is (get_bool_flag "PATTERN_CLAUSES_FORALL_AS_LIT"), unless we've temporarily changed it. ***)
 		new_pattern_clauses_2a nalllit m mr evars strict unmatched abslits
 	  | _ ->
 	      new_pattern_clauses_2a nalllit m mr evars strict unmatched abslits
@@ -699,31 +731,120 @@ let rec reflexivity_used_p m =
 
 let insane_instance_p m = (irrelevant_choice_p m) || (reflexivity_used_p m)
 
-let new_instantiation a m =
-  if (insane_instance_p m) then
+let seen_instantiations : ((stp * trm),unit) Hashtbl.t = Hashtbl.create 10
+
+let seen_inst_p a m =
+  Hashtbl.mem seen_instantiations (a,m)
+
+let see_inst a m =
+  Hashtbl.add seen_instantiations (a,m) ()
+
+let reset_search_1 () =
+  reset_search();
+  Hashtbl.clear seen_instantiations
+
+let rec stp_weight a =
+  match a with
+  | Ar(a1,a2) -> (get_int_flag "ARTP_WEIGHT") + (stp_weight a1) + (stp_weight a2)
+  | Base(_) -> get_int_flag "BASETP_WEIGHT"
+  | Prop -> get_int_flag "OTP_WEIGHT"
+
+let rec tm_weight m =
+  match m with
+  | Name(_,a) ->
+    let fac = get_int_flag "NAME_TP_WEIGHT_FAC" in
+    begin
+      if (fac > 0) then
+	(get_int_flag "NAME_WEIGHT") + (fac * (stp_weight a))
+      else
+	get_int_flag "NAME_WEIGHT"
+    end
+  | False -> get_int_flag "FALSE_WEIGHT"
+  | Imp -> (get_int_flag "IMP_WEIGHT")
+  | Forall(a) ->
+    let fac = get_int_flag "FORALL_TP_WEIGHT_FAC" in
+    begin
+      if (fac > 0) then
+	(get_int_flag "FORALL_WEIGHT") + (fac * (stp_weight a))
+      else
+	(get_int_flag "FORALL_WEIGHT")
+    end
+  | Eq(a) ->
+    let fac = get_int_flag "EQ_TP_WEIGHT_FAC" in
+    begin
+      if (fac > 0) then
+	(get_int_flag "EQ_WEIGHT") + (fac * (stp_weight a))
+      else
+	(get_int_flag "EQ_WEIGHT")
+    end
+  | Choice(a) ->
+    let fac = get_int_flag "CHOICE_TP_WEIGHT_FAC" in
+    begin
+      if (fac > 0) then
+	(get_int_flag "CHOICE_WEIGHT") + (fac * (stp_weight a))
+      else
+	(get_int_flag "CHOICE_WEIGHT")
+    end
+  | DB(i,a) ->
+    let fac = get_int_flag "DB_TP_WEIGHT_FAC" in
+    begin
+      if (fac > 0) then
+	(get_int_flag "DB_WEIGHT") + (fac * (stp_weight a))
+      else
+	(get_int_flag "DB_WEIGHT")
+    end
+  | Lam(a,m) ->
+    let fac = get_int_flag "LAM_TP_WEIGHT_FAC" in
+    begin
+      if (fac > 0) then
+	(get_int_flag "LAM_WEIGHT") + (fac * (stp_weight a))
+      else
+	(get_int_flag "LAM_WEIGHT")
+    end
+  | Ap(m1,m2) ->
+    (get_int_flag "AP_WEIGHT") + (tm_weight m1) + (tm_weight m2)
+  | _ -> 0
+
+let instantiation_priority a m =
+  (get_int_flag "INSTANTIATION_DELAY") + (stp_weight a) + (tm_weight m)
+
+let process_instantiation a m =
+  begin (*** Process it now -- making it passive ***)
+    if (!verbosity > 5) then (Printf.printf "Instantiating with %s\n" (trm_str m));
+    add_instantiation a m;
+    List.iter
+      (fun (lit,p) -> (*** Instantiate known universal predicates ***)
+	if (not (filteredp (-lit))) then
+	  let pm = (norm name_def (ap(p,m))) in
+	  insertWithPriority (get_int_flag "FORALL_DELAY") (ProcessProp1 pm);
+	  let pmlit = get_literal pm in
+	  new_search_clause [lit;pmlit]
+	    (if (!mkproofterm) then (Some(InstRule(a,p,m))) else None)
+	    )
+      (Hashtbl.find_all univpreds a)
+  end
+
+let new_active_instantiation a m =
+  if (insane_instance_p m) then (*** Always leave insane instantiations active with a very high priority when we first see them. ***)
     begin
       if (!verbosity > 40) then (Printf.printf "Delaying insane instantiation %s\n" (trm_str m));
-      insertWithPriority 5000 (NewInst(a,m)) (*** Delay it for a very long time ***)
+      insertWithPriority 5000 (NewInst(a,m)) (*** Delay it for a very long time, leaving it active ***)
     end
-  else
+  else if (get_bool_flag "EAGERLY_PROCESS_INSTANTIATIONS") then (*** Process it now -- making it passive ***)
+    process_instantiation a m
+  else (*** Determine priority and put it on the queue (active instantiation) ***)
+    let d = instantiation_priority a m in
     begin
-      if (!verbosity > 5) then (Printf.printf "Instantiating with %s\n" (trm_str m));
-      add_instantiation a m;
-      List.iter
-	(fun (lit,p) -> (*** Instantiate known universal predicates ***)
-	  if (not (filteredp (-lit))) then
-	    let pm = (norm name_def (ap(p,m))) in
-	    insertWithPriority (get_int_flag "FORALL_DELAY") (ProcessProp1 pm);
-	    let pmlit = get_literal pm in
-	    new_search_clause [lit;pmlit]
-	      (if (!mkproofterm) then (Some(InstRule(a,p,m))) else None)
-	)
-	(Hashtbl.find_all univpreds a)
+      if (!verbosity > 20) then (Printf.printf "Delaying with priority %d active instantiation %s\n" d (trm_str m));
+      insertWithPriority d (NewInst(a,m))
     end
 
 let possibly_new_instantiation a m =
-  if (not (known_instantiation a m)) then
-    new_instantiation a m
+  if (not (seen_inst_p a m)) then
+    begin
+      see_inst a m;
+      new_active_instantiation a m
+    end
 
 let decompose lit largs rargs r =
   let delay = get_int_flag "POST_DEC_DELAY" in
@@ -911,7 +1032,7 @@ let is_an_eqn m =
 
 let rec leibeq_to_primeq m =
   try
-    let (a,s,t) = is_an_eqn m in eq a s t
+    let (a,s,t) = is_an_eqn m in eq a (leibeq_to_primeq s) (leibeq_to_primeq t) (*** added the recursive calls here so it hopefully behaves like Andreas Teucke's version in proofterm.  - Chad, Feb 15, 2011 ***)
   with
   | Not_found ->
       begin
@@ -957,7 +1078,7 @@ let rec enum_depth d ctx a =
       if (d > 0) then
 	begin
 	  let r = ref [] in
-	  enum_depth_proj (d - 1) ctx ctx a 0 r;  (*** - 1 here ensures termination ***)
+	  ignore (enum_depth_proj (d - 1) ctx ctx a 0 r);  (*** - 1 here ensures termination ***)
 	  enum_depth_imit (d - 1) ctx a r; (*** - 1 here ensures termination ***)
 	  (!r)
 	end
@@ -1018,8 +1139,10 @@ and enum_depth_spine d ctx sigmal h r =
 	(enum_depth d ctx a)
   | [] -> r := (h::!r)
 
+let already_preprocessed = ref false
+
 let preprocess m =
-  if (get_bool_flag "LEIBEQ_TO_PRIMEQ") then
+  if ((get_bool_flag "LEIBEQ_TO_PRIMEQ") && (not (!already_preprocessed))) then
     norm name_def (leibeq_to_primeq m)
   else
     m
@@ -1079,7 +1202,7 @@ let process_search_option op =
 			      match a with
 			      | Ar(a1,a2) -> (*** FE ***)
 (***				  let m3 = neg (forall a1 (eq a2 (norm name_def (Ap (shift m1 0 1,DB (0,a1)))) (norm name_def (Ap (shift m2 0 1,DB (0,a1)))))) in  ***)
-				  let m3 = neg (forall a1 (eq a2 (norm name_def (Ap (m1,DB (0,a1)))) (norm name_def (Ap (m2,DB (0,a1)))))) in (*** No need to shift, since no dangling DB's [Thanks to Teucke!] ***)
+				  let m3 = norm name_def (neg (forall a1 (eq a2 (Ap (m1,DB (0,a1))) (Ap (m2,DB (0,a1)))))) in (*** No need to shift, since no dangling DB's [Thanks to Teucke!] ; Mar 15, normalize at the end to handle final eta (see FQ too) ***)
 				  insertWithPriority (get_int_flag "POST_FEQ_DELAY") (ProcessProp1 m3);
 				  if (get_bool_flag "INSTANTIATE_WITH_FUNC_DISEQN_SIDES") then
 				    begin
@@ -1161,7 +1284,7 @@ let process_search_option op =
 		end
 	    | Ap(Forall(a),m1) -> (*** Instantiate ***)
 		begin
-		  if (get_bool_flag "ENABLE_PATTERN_CLAUSES") then (apply_pattern_clauses mlit m; new_pattern_clauses (-mlit) m);
+		  if (get_bool_flag "ENABLE_PATTERN_CLAUSES") then (apply_pattern_clauses mlit m; if (get_bool_flag "DYNAMIC_PATTERN_CLAUSES") then new_pattern_clauses (-mlit) m);
 		  if ((get_bool_flag "FILTER_UNIV") && (filterp mlit)) then
 		    ()
 		  else
@@ -1208,7 +1331,7 @@ let process_search_option op =
 					    (fun beta _ -> insertWithPriority (d + d_i) (EnumTp(d_i,Base(beta),Base(beta))))
 					    name_base;
 					  Hashtbl.iter
-					    (fun x (m,sigma) ->
+					    (fun x ((m,sigma),_) ->
 					      let (sigmal,rtp) = argtps_rtp sigma in
 					      if (Hashtbl.mem name_def x) then
 						begin
@@ -1281,11 +1404,25 @@ let process_search_option op =
 			match args with
 			  [m1;m2] ->
 			    begin
+			      if ((get_bool_flag "PATTERN_CLAUSES_TRANSITIVITY_EQ") && (not (Hashtbl.mem pattern_clauses_transitivity_types a))) then
+				begin (*** April 6, 2011 - add a pattern clause for transitivity of equality the first time we see an equation at a type. ***)
+				  let transa = forall a (forall a (forall a (imp (eq a (DB(2,a)) (DB(1,a))) (imp (eq a (DB(1,a)) (DB(0,a))) (eq a (DB(2,a)) (DB(0,a))))))) in
+				  let transa2 = forall a (forall a (forall a (imp (eq a (DB(2,a)) (DB(1,a))) (imp (eq a (DB(0,a)) (DB(1,a))) (eq a (DB(2,a)) (DB(0,a))))))) in
+				  let transa3 = forall a (forall a (forall a (imp (eq a (DB(1,a)) (DB(2,a))) (imp (eq a (DB(1,a)) (DB(0,a))) (eq a (DB(2,a)) (DB(0,a))))))) in
+				  Hashtbl.add pattern_clauses_transitivity_types a ();
+				  pattern_clauses_forall_as_lit := false;
+				  pattern_clauses_onlyallstrict := false;
+				  new_pattern_clauses (- (get_literal transa)) transa;
+				  new_pattern_clauses (- (get_literal transa2)) transa2;
+				  new_pattern_clauses (- (get_literal transa3)) transa3;
+				  pattern_clauses_forall_as_lit := get_bool_flag "PATTERN_CLAUSES_FORALL_AS_LIT";
+				  pattern_clauses_onlyallstrict := get_bool_flag "PATTERN_CLAUSES_ONLYALLSTRICT";
+				end;
 			      match a with
 			      | Ar(a1,a2) -> (*** FQ ***)
 				  if (get_bool_flag "ENABLE_PATTERN_CLAUSES") then (apply_pattern_clauses mlit m);
 (***				  let m3 = forall a1 (eq a2 (norm name_def (Ap (shift m1 0 1,DB (0,a1)))) (norm name_def (Ap (shift m2 0 1,DB (0,a1))))) in ***)
-				  let m3 = forall a1 (eq a2 (norm name_def (Ap (m1,DB (0,a1)))) (norm name_def (Ap (m2,DB (0,a1))))) in
+				  let m3 = norm name_def (forall a1 (eq a2 (Ap (m1,DB (0,a1))) (Ap (m2,DB (0,a1))))) in (*** See FE above for why there's no shift on m1 and m2. Mar 15 2011: Normalize after everything, to make sure the case of a final eta is handled, e.g., forall x.s=x where s has no x free eta reduces. ***)
 				  insertWithPriority (get_int_flag "POST_NFEQ_DELAY") (ProcessProp1 m3);
 				  let m3lit = get_literal m3 in
 				  new_search_clause [-mlit;m3lit] r
@@ -1375,11 +1512,11 @@ let process_search_option op =
 	match insts with
 	| [] -> 
 	    let m = default_elt aname in
-	    new_instantiation a m
+	    possibly_new_instantiation a m
 	| _ -> ()
       end
   | NewInst(a,m) ->
-      possibly_new_instantiation a m;
+      process_instantiation a m; (*** Must process it now and make it passive ***)
   | EnumIterDeep(d,a) ->
       for i = (!typesgenerated) to (d - 1)
       do
@@ -1515,19 +1652,111 @@ let search_main () =
       | Not_found -> raise (GenericError "Not_found raised while processing search option")
     done
   with Not_found -> raise Satisfiable
-      
-let search_1 b =
+
+let pos_heads : (trm,unit) Hashtbl.t = Hashtbl.create 10
+let neg_heads : (trm,unit) Hashtbl.t = Hashtbl.create 10
+let eqn_heads : (trm,unit) Hashtbl.t = Hashtbl.create 10
+let diseqn_heads : (trm,unit) Hashtbl.t = Hashtbl.create 10
+
+let rec set_relevance_info_tm m =
+  match m with
+  | Lam(a,m1) -> set_relevance_info_tm m1
+  | _ ->
+      match (head_spine m) with
+      | (Name(x,_) as h,s) ->
+	  Hashtbl.add diseqn_heads h ();
+	  ignore (List.map set_relevance_info_tm s)
+      |	_ -> ()
+
+let rec set_relevance_info m pos =
+  match m with
+  | Lam(a,m1) -> set_relevance_info m1 pos
+  | _ ->
+      match (head_spine m) with
+      | (Name(x,_) as h,s) ->
+	  ignore (List.map set_relevance_info_tm s);
+	  if pos then
+	    Hashtbl.add pos_heads h ()
+	  else
+	    Hashtbl.add neg_heads h ()
+      | (Eq(_),[m1;m2]) ->
+	  begin
+	    match (head_spine m1) with
+	    | (Name(x,_) as h,s) ->
+		if pos then
+		  Hashtbl.add eqn_heads h ()
+		else
+		  (Hashtbl.add diseqn_heads h ();
+		   ignore (List.map set_relevance_info_tm s))
+	    |  _ -> ();
+	    match (head_spine m2) with
+	    | (Name(x,_) as h,s) ->
+		if pos then
+		  Hashtbl.add eqn_heads h ()
+		else
+		  (Hashtbl.add diseqn_heads h ();
+		   ignore (List.map set_relevance_info_tm s))
+	    |  _ -> ()
+	  end
+      | (Imp,[m1;m2]) ->
+	  begin
+	    set_relevance_info m1 (not pos);
+	    set_relevance_info m2 pos
+	  end
+      | (Forall(_),[m1]) -> set_relevance_info m1 pos
+      | (_,_) -> ()
+
+let rec relevance_delay m pos d =
+  if (d <= 0) then
+    0
+  else
+    match m with
+    | Lam(a,m1) -> relevance_delay m1 pos d
+    | _ ->
+	match (head_spine m) with
+	| (Name(x,_) as h,s) ->
+	    if ((pos && (Hashtbl.mem neg_heads h)) || ((not pos) && (Hashtbl.mem pos_heads h))) then
+	      d - 1
+	    else
+	      d
+	| (Eq(_),[m1;m2]) ->
+	    begin
+	      let d2 =
+		match (head_spine m1) with
+		| (Name(x,_) as h,s) ->
+		    if ((pos && (Hashtbl.mem diseqn_heads h)) || ((not pos) && (Hashtbl.mem eqn_heads h))) then
+		      d - 1
+		    else
+		      d
+		| _ -> d
+	      in
+		match (head_spine m2) with
+		| (Name(x,_) as h,s) ->
+		    if ((pos && (Hashtbl.mem diseqn_heads h)) || ((not pos) && (Hashtbl.mem eqn_heads h))) then
+		      d2 - 1
+		    else
+		      d2
+		| _ -> d2
+	    end
+	| (Imp,[m1;m2]) -> relevance_delay m1 (not pos) (relevance_delay m2 pos d)
+	| (Forall(_),[m1]) -> relevance_delay m1 pos d
+	| (_,_) -> d
+
+let search_1 b1 b2 =
   begin
     let _ = minisat_init (!verbosity) in
     if (!verbosity > 3) then Printf.printf "Initialized minisat\n";
+    (*** Process "conjecture" part first - March 18 2011 ***)
     List.iter (fun m ->
       let mn = preprocess m in
       match (choiceop_axiom mn) with
       | Some(x,a) ->
-	  declare_choiceop x a
+	  declare_choiceop x a mn
       | None ->
 	  begin
-	   insertWithPriority 0 (ProcessProp1 mn);
+	    insertWithPriority 0 (ProcessProp1 mn);
+	    if ((get_int_flag "RELEVANCE_DELAY") > 0) then
+	      set_relevance_info mn false;
 	    let mlit = get_literal mn in
 	    new_assumption_clause [mlit];
 	    if (get_bool_flag "INITIAL_SUBTERMS_AS_INSTANTIATIONS") then
@@ -1539,7 +1768,30 @@ let search_1 b =
 		if (!verbosity > 4) then Printf.printf "---- Initial Subterms as Instantiations END ----\n"
 	      end
 	  end)
-      b;
+      b1;
+    (*** Process "axiom" part second - March 18 2011 ***)
+    List.iter (fun m ->
+      let mn = preprocess m in
+      match (choiceop_axiom mn) with
+      | Some(x,a) ->
+	  declare_choiceop x a mn
+      | None ->
+	  begin
+	    insertWithPriority
+		((get_int_flag "AXIOM_DELAY") + (relevance_delay mn true (get_int_flag "RELEVANCE_DELAY")))
+	      (ProcessProp1 mn);
+	    let mlit = get_literal mn in
+	    new_assumption_clause [mlit];
+	    if (get_bool_flag "INITIAL_SUBTERMS_AS_INSTANTIATIONS") then
+	      begin
+		if (!verbosity > 4) then Printf.printf "---- Initial Subterms as Instantiations BEGIN ----\n";
+		if (!verbosity > 4) then Printf.printf "---- m = %s ----\n" (trm_str m);
+		if (!verbosity > 4) then Printf.printf "---- mn = %s ----\n" (trm_str mn);
+		ignore (subterms_as_instantiations mn 0);
+		if (!verbosity > 4) then Printf.printf "---- Initial Subterms as Instantiations END ----\n"
+	      end
+	  end)
+      b2;
     begin
       if ((get_bool_flag "FILTER_UNIV_USABLE") || (get_bool_flag "FILTER_POSATM_USABLE") || (get_bool_flag "FILTER_NEGATM_USABLE") || (get_bool_flag "FILTER_POSEQ_USABLE") || (get_bool_flag "FILTER_NEGEQ_USABLE")) then
 	let d = (get_int_flag "FILTER_START") in
@@ -1548,8 +1800,8 @@ let search_1 b =
     search_main ()
   end
 
-let rec split_global_disjunctions p b sgn =
-  match p with
+let rec split_global_disjunctions b1 b2 b bs sgn cnj =
+  match b1 with
     (m::pr) ->
       begin
 	match m with
@@ -1568,7 +1820,7 @@ let rec split_global_disjunctions p b sgn =
 	      let nm1w = (normneg m1w) in
 	      begin
 		try
-		  split_global_disjunctions (nm1w::pr) b sgn
+		  split_global_disjunctions (nm1w::pr) b2 b bs sgn cnj
 		with
 		| Unsatisfiable(Some(r1)) ->
 		    raise (Unsatisfiable(Some(NegAllR(a,m1,ws,r1))))
@@ -1578,14 +1830,14 @@ let rec split_global_disjunctions p b sgn =
 	    let (h,args) = head_spine nm in
 	    begin
 	      match h with
-	      | False -> split_global_disjunctions pr b sgn (*** drop ~False ***)
+	      | False -> split_global_disjunctions pr b2 b bs sgn cnj (*** drop ~False ***)
 	      | Imp ->
 		  begin
 		    match args with
 		      [m1;m2] ->
 			begin
 			  try
-			    split_global_disjunctions (m1::(normneg m2)::pr) b sgn
+			    split_global_disjunctions (m1::(normneg m2)::pr) b2 b bs sgn cnj
 			  with
 			  | Unsatisfiable(Some(r1)) ->
 			      raise (Unsatisfiable (Some (NegImpR(m1,m2,r1))))
@@ -1602,7 +1854,7 @@ let rec split_global_disjunctions p b sgn =
 			  try
 			    split_global_disjunctions
 			      ((norm name_def (normneg (forall a1 (eq a2 (ap(m1s,DB(0,a1))) (ap(m2s,DB(0,a1)))))))::pr)
-			      b sgn
+			      b2 b bs sgn cnj
 			  with
 			  | Unsatisfiable(Some(r1)) ->
 			      raise (Unsatisfiable (Some (NegEqFR(a1,a2,m1,m2,r1))))
@@ -1615,25 +1867,25 @@ let rec split_global_disjunctions p b sgn =
 		      [m1;m2] ->
 			begin
 			  try
-			    split_global_disjunctions (m1::(normneg m2)::pr) b sgn
+			    split_global_disjunctions (m1::(normneg m2)::pr) b2 b bs sgn cnj
 			  with
 			  | Unsatisfiable(Some(r1)) ->
 			      begin
 				try
-				  reset_search (); incr sgn;
-				  split_global_disjunctions ((normneg m1)::m2::pr) b sgn
+				  reset_search_1 (); incr sgn;
+				  split_global_disjunctions ((normneg m1)::m2::pr) b2 b bs sgn cnj
 				with
 				| Unsatisfiable(Some(r2)) ->
 				    raise (Unsatisfiable (Some (NegEqOR(m1,m2,r1,r2))))
 			      end
 			  | Unsatisfiable(None) ->
-			      reset_search (); incr sgn;
-			      split_global_disjunctions ((normneg m1)::m2::pr) b sgn
+			      reset_search_1 (); incr sgn;
+			      split_global_disjunctions ((normneg m1)::m2::pr) b2 b bs sgn cnj
 			end
 		    | _ -> raise (GenericError "foo3")
 		  end
 	      | _ ->
-		  split_global_disjunctions pr (m::b) sgn
+		  split_global_disjunctions pr b2 (m::b) bs sgn cnj
 	    end
 	| _ ->
 	    let (h,args) = head_spine m in
@@ -1646,20 +1898,20 @@ let rec split_global_disjunctions p b sgn =
 		      [m1;m2] ->
 			begin
 			  try
-			    split_global_disjunctions ((normneg m1)::pr) b sgn
+			    split_global_disjunctions ((normneg m1)::pr) b2 b bs sgn cnj
 			  with
 			  | Unsatisfiable(Some(r1)) ->
 			      begin
 				try
-				  reset_search (); incr sgn;
-				  split_global_disjunctions (m2::pr) b sgn
+				  reset_search_1 (); incr sgn;
+				  split_global_disjunctions (m2::pr) b2 b bs sgn cnj
 				with
 				| Unsatisfiable(Some(r2)) ->
 				    raise (Unsatisfiable (Some (ImpR(m1,m2,r1,r2))))
 			      end
 			  | Unsatisfiable(None) ->
-			      reset_search (); incr sgn;
-			      split_global_disjunctions (m2::pr) b sgn
+			      reset_search_1 (); incr sgn;
+			      split_global_disjunctions (m2::pr) b2 b bs sgn cnj
 			end
 		    | _ -> raise (GenericError "foo3")
 		  end
@@ -1669,39 +1921,62 @@ let rec split_global_disjunctions p b sgn =
 		      [m1;m2] ->
 			begin
 			  try
-			    split_global_disjunctions (m1::m2::pr) b sgn
+			    split_global_disjunctions (m1::m2::pr) b2 b bs sgn cnj
 			  with
 			  | Unsatisfiable(Some(r1)) ->
 			      begin
 				try
-				  reset_search (); incr sgn;
-				  split_global_disjunctions ((normneg m1)::(normneg m2)::pr) b sgn
+				  reset_search_1 (); incr sgn;
+				  split_global_disjunctions ((normneg m1)::(normneg m2)::pr) b2 b bs sgn cnj
 				with
 				| Unsatisfiable(Some(r2)) ->
 				    raise (Unsatisfiable (Some (EqOR(m1,m2,r1,r2))))
 			      end
 			  | Unsatisfiable(None) ->
-			      reset_search (); incr sgn;
-			      split_global_disjunctions ((normneg m1)::(normneg m2)::pr) b sgn
+			      reset_search_1 (); incr sgn;
+			      split_global_disjunctions ((normneg m1)::(normneg m2)::pr) b2 b bs sgn cnj
 			end
 		    | _ -> raise (GenericError "foo3")
 		  end
 	      | _ ->
-		  split_global_disjunctions pr (m::b) sgn
+		  split_global_disjunctions pr b2 (m::b) bs sgn cnj
 	    end
       end
   | [] ->
       begin
-	if (!verbosity > 5) then Printf.printf "Searching on Subgoal %d\n" (!sgn);
-	search_1 b;
+	match b2 with
+	  (_::_) -> split_global_disjunctions b2 [] [] b sgn false
+	| [] ->
+	    begin
+	      if (!verbosity > 5) then Printf.printf "Searching on Subgoal %d\n" (!sgn);
+	      if cnj then
+		search_1 b bs
+	      else
+		search_1 bs b (*** reversed conjecture and axiom parts to compute axiom parts ***)
+	    end
       end
 	
 let search () =
   begin
     search_init();
-    if (get_bool_flag "SPLIT_GLOBAL_DISJUNCTIONS") then
-      split_global_disjunctions (!initial_branch) [] (ref 1) (*** could I preprocess first? - Feb 1, 2011 ***)
+    pattern_clauses_forall_as_lit := get_bool_flag "PATTERN_CLAUSES_FORALL_AS_LIT"; (*** Chad: April 6, 2011 ***)
+    pattern_clauses_onlyallstrict := get_bool_flag "PATTERN_CLAUSES_ONLYALLSTRICT"; (*** Chad: April 6, 2011 ***)
+    pattern_clauses_eqnlits := get_bool_flag "PATTERN_CLAUSES_EQNLITS"; (*** Chad: April 6, 2011 ***)
+    let b1 = ref [] in
+    let b2 = ref [] in
+    if (get_bool_flag "TREAT_CONJECTURE_AS_SPECIAL") then
+      ignore (List.map (fun m -> if (Hashtbl.mem part_of_conjecture m) then b1 := (m::!b1) else b2 := (m::!b2)) (!initial_branch)) (*** March 18 2011; split branch into conjecture part (b1) and axiom part (b2) ***)
     else
-      search_1 (!initial_branch)
+      ignore (b2 := (!initial_branch)); (*** March 20 2011: In this case it should behave essentially as Satallax 2.0 did. ***)
+    if (get_bool_flag "SPLIT_GLOBAL_DISJUNCTIONS") then
+      if (get_bool_flag "PREPROCESS_BEFORE_SPLIT") then
+	let pbr1 = (List.map preprocess (!b1)) in
+	let pbr2 = (List.map preprocess (!b2)) in
+	already_preprocessed := true;
+	split_global_disjunctions pbr1 pbr2 [] [] (ref 1) true (*** Following Andreas Teucke's suggestion: Preprocess before split - Mar 18 2011 ***)
+      else
+	split_global_disjunctions (!b1) (!b2) [] [] (ref 1) true
+    else
+      search_1 (!b1) (!b2)
   end
 
