@@ -11,8 +11,8 @@ exception Redeclaration of string
 exception FileNotFound of string
 exception GenericError of string
 
-let mode : string option ref = ref None
-let timeout : int option ref = ref None
+let mode : string list ref = ref []
+let timeout : float option ref = ref None
 let verbosity : int ref = ref 1 (*** Default to 1; 0 should mean print nothing (-verb 0 command line option) ***)
 let coq = ref false
 let coqchannel : out_channel option ref = ref None
@@ -28,7 +28,7 @@ let coqsig_def_trm : (string * trm) list ref = ref []
 let coqsig_hyp_trm : (string * trm) list ref = ref []
 let name_base : (string,unit) Hashtbl.t = Hashtbl.create 10
 let name_tp : (string,stp) Hashtbl.t = Hashtbl.create 100
-let name_trm : (string,trm * stp) Hashtbl.t = Hashtbl.create 100
+let name_trm : (string,(trm * stp) * bool ref) Hashtbl.t = Hashtbl.create 100
 let name_def : (string,trm) Hashtbl.t = Hashtbl.create 100
 let name_def_prenorm : (string,trm) Hashtbl.t = Hashtbl.create 100
 let name_hyp : (string,trm) Hashtbl.t = Hashtbl.create 100
@@ -36,18 +36,25 @@ let coq_used_names : (string,unit) Hashtbl.t = Hashtbl.create 100
 let coq_names : (string,string) Hashtbl.t = Hashtbl.create 100
 let coq_hyp_names : (string,string) Hashtbl.t = Hashtbl.create 100
 
+(*** March 31, 2011 - Chad - THF policy regarding definitions. (See comments before declare_definition_real below. ***)
+exception TreatAsAssumption
+
 let next_fresh_name : int ref = ref 0
 
 let rec get_fresh_name a =
   let x = "__" ^ (string_of_int (!next_fresh_name)) in
   incr next_fresh_name;
+  if (!coq) then
+    begin
+      ignore (coqify_name x coq_names coq_used_names)
+    end;
   if ((Hashtbl.mem name_base x) || (Hashtbl.mem name_tp x)) then
     get_fresh_name a
   else
     let xa = Name(x,a) in
     begin
       Hashtbl.add name_tp x a;
-      Hashtbl.add name_trm x (xa,a);
+      Hashtbl.add name_trm x ((xa,a), ref false);
       (x,xa)
     end
 
@@ -90,12 +97,7 @@ let get_atom m =
 let get_literal m =
   match (neg_body m) with
     Some n ->
-      begin
-	match (neg_body n) with
-	| Some _ -> raise (GenericError("DN Bug - get_literal called with dneg " ^ (trm_str m)))
-	| None ->
-	    (- (get_atom n))
-      end
+      (- (get_atom n))
   | None -> (get_atom m)
 let literal_to_trm l =
   if (l < 0) then
@@ -110,6 +112,8 @@ type clause = int list
 
 let clauses : clause list ref = ref []
 
+let clausesTable : (clause,unit) Hashtbl.t = Hashtbl.create 10
+
 type searchoption =
  (*** ProcessProp1(m) For rules requiring one formula in the head (most rules) ***)
     ProcessProp1 of trm
@@ -179,7 +183,7 @@ let searchoption_str s =
   | Enum(d,gamma,sigma,_) -> "Enum(" ^ (string_of_int d) ^ "," ^ (String.concat ";" (List.map stp_str gamma)) ^ "," ^ (stp_str sigma) ^ ")"
   | NewInst(_,m) -> "NewInst(" ^ (trm_str m) ^ ")"
   | Filter(d) -> "Filter(" ^ (string_of_int d) ^ ")"
-  | _ -> "OtherSearchOption" (*** to do: the rest if we need them ***)
+(***  | _ -> "OtherSearchOption" (*** to do: the rest if we need them ***) ***)
 
 (*** For printing a ruleinfo - helpful for debugging. - Chad, Oct 2010 ***)
 let ruleinfo_str s =
@@ -222,6 +226,8 @@ let rec print_refut r =
       Printf.printf "NegEqOR\n  left: %s\n  right: %s\n" (trm_str m) (trm_str n);
       print_refut r1;
       print_refut r2
+  | AssumptionConflictR ->
+      Printf.printf "AssumptionConflictR\n";
   | SearchR(cl,cr) ->
       Printf.printf "SearchR\n  clauses with search options:\n";
       List.iter
@@ -241,28 +247,47 @@ let rec print_refut r =
 let processed : (trm,unit) Hashtbl.t = Hashtbl.create 100
 let clause_ruleinfo : (clause,ruleinfo) Hashtbl.t ref = ref (Hashtbl.create 100)
 
-let insert_assumption_clause c = clauses := (c::!clauses)
+exception DuplicateClause
+
+let insert_assumption_clause c =
+  if (Hashtbl.mem clausesTable c) then
+    raise DuplicateClause
+  else
+    begin
+      Hashtbl.add clausesTable c ();
+      clauses := (c::!clauses)
+    end
 
 let insert_search_clause c r = 
-  begin
-    clauses := (c::!clauses);
-    match r with
-    | Some(r) ->
-	Hashtbl.add (!clause_ruleinfo) c r;
-    | None -> ()
-  end
+(***  (Printf.printf "inserting search clause %s: %s\n" (List.fold_left (fun x y -> if ((String.length x) == 0) then (string_of_int y) else (x ^ " " ^ (string_of_int y))) "" c) (match r with (Some r) -> (ruleinfo_str r) | _ -> "None"); flush stdout); ***)
+  if (Hashtbl.mem clausesTable c) then
+    (*** Chad: April 4, 2011: Check if the clause has already been added before.  If it has, raise DuplicateClause. (Otherwise, we may end up with bad rule info -- e.g., a later rule which will violate freshness) ***)
+(***    Printf.printf "duplicate clause %s: %s\n" (List.fold_left (fun x y -> if ((String.length x) == 0) then (string_of_int y) else (x ^ " " ^ (string_of_int y))) "" c) (match r with (Some r) -> (ruleinfo_str r) | _ -> "None"); flush stdout; ***)
+     raise DuplicateClause
+  else
+    begin
+      Hashtbl.add clausesTable c ();
+      clauses := (c::!clauses);
+      match r with
+      | Some(r) ->
+	  Hashtbl.add (!clause_ruleinfo) c r;
+      | None -> ()
+    end
 
 (*** Different Implementations of Fair Priority Queues ***)
 module PQueue0 = PriorityQueue0 (struct type t = searchoption end)
 
 module PQueue1 = PriorityQueue1 (struct type t = searchoption end)
 
+module PQueue2 = PriorityQueue2 (struct type t = searchoption end)
+
 let insertWithPriority n o =
   let (ins,dp) =
     begin
       match (get_int_flag "PRIORITY_QUEUE_IMPL") with
       | 0 -> (PQueue0.insertWithPriority,PQueue0.debug_print)
       | 1 -> (PQueue1.insertWithPriority,PQueue1.debug_print)
+      | 2 -> (PQueue2.insertWithPriority,PQueue2.debug_print)
       | n -> (raise (GenericError("Invalid value of PRIORITY_QUEUE_IMPL: " ^ (string_of_int n))))
     end
   in
@@ -278,6 +303,7 @@ let getNext () =
       match (get_int_flag "PRIORITY_QUEUE_IMPL") with
       | 0 -> (PQueue0.getNext,PQueue0.debug_print)
       | 1 -> (PQueue1.getNext,PQueue1.debug_print)
+      | 2 -> (PQueue2.getNext,PQueue2.debug_print)
       | n -> (raise (GenericError("Invalid value of PRIORITY_QUEUE_IMPL: " ^ (string_of_int n))))
     end
   in
@@ -292,6 +318,7 @@ let peekAtNext() =
     match (get_int_flag "PRIORITY_QUEUE_IMPL") with
     | 0 -> PQueue0.peekAtNext()
     | 1 -> PQueue1.peekAtNext()
+    | 2 -> PQueue2.peekAtNext()
     | n -> (raise (GenericError("Invalid value of PRIORITY_QUEUE_IMPL: " ^ (string_of_int n))))
   end
 
@@ -379,7 +406,7 @@ let add_instantiation a m =
     Hashtbl.add h m ();
     set_default_elt_if_needed a m
 
-let choiceopnames : (string,stp) Hashtbl.t = Hashtbl.create 10
+let choiceopnames : (string,(stp * trm)) Hashtbl.t = Hashtbl.create 10
 
 let choiceop_axiom m =
   match m with
@@ -409,7 +436,7 @@ let choiceop_axiom m =
     -> Some(x,a)
   | _ -> None
 
-let declare_choiceop x a = Hashtbl.add choiceopnames x a
+let declare_choiceop x a m = Hashtbl.add choiceopnames x (a,m)
 
 let choiceop m =
   match m with
@@ -417,7 +444,8 @@ let choiceop m =
   | Name(x,_) ->
       begin
 	try
-	  Some(Hashtbl.find choiceopnames x)
+	  let (a,_) = Hashtbl.find choiceopnames x in
+	  Some(a)
 	with
 	| Not_found -> None
       end
@@ -425,6 +453,8 @@ let choiceop m =
 
 let filtered : (int,unit) Hashtbl.t = Hashtbl.create 10
 
+let part_of_conjecture : (trm,unit) Hashtbl.t = Hashtbl.create 10
+
 type namecategory =
     ChoiceOp of int * int * stp list * trm list (*** (i,n,sigmal,tl) where length of sigmal and tl are n, 0 <= i < n, Name has type (sigmal -> o) -> sigmal[i], and for each j:{0,...,n-1} tl[j] is the jth component of the n-ary choice operator (in particular, tl[i] is this one) ***)
    | DescrOp of int * int * stp list * trm list (*** (i,n,sigmal,tl) where length of sigmal and tl are n, 0 <= i < n, Name has type (sigmal -> o) -> sigmal[i], and for each j:{0,...,n-1} tl[j] is the jth component of the n-ary description operator (in particular, tl[i] is this one) ***)
@@ -674,31 +704,57 @@ let declare_typed_constant (name:string) (role:string) (m:pretrm) =
 		  coqsig_const := (y,tp)::!coqsig_const
 		end;
 	      Hashtbl.add name_tp x tp;
-	      Hashtbl.add name_trm x (Name(x,tp),tp)
+	      Hashtbl.add name_trm x ((Name(x,tp),tp),ref false)
 	end
     | _ -> raise (GenericError ("Incorrect format for type declaration " ^ name))
   end
     
+(***
+ THF official policy requires that all names are given a type before use,
+ that definitions can be circular (in which case they must be treated as equations),
+ and that definitions may appear *after* it's already been used in an assumption.
+ In order to comply with this policy *and* do something reasonable in the 'usual' case,
+ I do the following:
+
+ 1. I keep a hashtable (occurred_names) of typed names that have been used in an assumption.
+ 2. When I encounter a definition, if it's of the form (x = t), I first parse t.  If x has been used,
+    then I include the equation as an axiom.  Otherwise, x is treated as a definition and expanded away.
+
+ NOTE: If I wanted to be strict, I would exit if the definition is given without the name being given a type,
+ but I will not require the type to be given.  I will print a warning if the verbosity is > 0.
+ ***)
 let declare_definition_real x m =
   if (Hashtbl.mem name_base x) then raise (Redeclaration x);
-  if (Hashtbl.mem name_def x) then raise (Redeclaration x);
+  if (Hashtbl.mem name_def x) then raise TreatAsAssumption; (*** treat it as an assumption per THF policy. - Mar 31, 2011 - Chad ***)
+(*** raise (Redeclaration x); ***)
   try
     let tp = Hashtbl.find name_tp x in
     let tm = st_to_trm_given_stp m tp in
-    if (!coq) then
-      begin (*** The name was coqified when the type was declared. ***)
-	try
-	  let y = Hashtbl.find coq_names x in
-	  coqsig_def := (y,m)::!coqsig_def;
-	  coqsig_def_trm := (y,tm)::!coqsig_def_trm
-	with
-	| Not_found -> raise (GenericError("Could not find Coq version of name " ^ x))
-      end;
-    Hashtbl.add name_def_prenorm x tm;
-    Hashtbl.add name_def x (norm name_def (logicnorm tm))
+    try 
+      let (_,r) = Hashtbl.find name_trm x in
+      if (!r) then
+	raise TreatAsAssumption (*** It's already been used, treat 'definition' as an assumption. ***)
+      else
+	raise Not_found
+    with
+    | Not_found ->
+	begin
+	  if (!coq) then
+	    begin (*** The name was coqified when the type was declared. ***)
+	      try
+		let y = Hashtbl.find coq_names x in
+		coqsig_def := (y,m)::!coqsig_def;
+		coqsig_def_trm := (y,tm)::!coqsig_def_trm
+	      with
+	      | Not_found -> raise (GenericError("Could not find Coq version of name " ^ x))
+	    end;
+	  Hashtbl.add name_def_prenorm x tm;
+	  Hashtbl.add name_def x (norm name_def (logicnorm tm))
+	end
   with
   | Not_found ->
-      begin
+      begin (*** Giving a definition without giving it's type isn't allowed in THF anymore.  I'm still allowing it.  ***)
+	if (!verbosity > 0) then Printf.printf "WARNING: %s defined without giving type first.\n" x;
 	let (tm,tp) = st_to_trm m in
 	if (!coq) then
 	  begin
@@ -708,23 +764,12 @@ let declare_definition_real x m =
 	    coqsig_def_trm := (y,tm)::!coqsig_def_trm
 	  end;
 	Hashtbl.add name_tp x tp;
-	Hashtbl.add name_trm x (Name(x,tp),tp);
+	Hashtbl.add name_trm x ((Name(x,tp),tp),ref false);
 	Hashtbl.add name_def_prenorm x tm;
 	Hashtbl.add name_def x (norm name_def (logicnorm tm))
       end
-	
-let declare_definition (name:string) (role:string) (m:pretrm) =
-  begin
-    if !verbosity > 4 then Printf.printf "declare_definition %s %s\n" name role;
-    match m with
-      PDef(PName(x),m) ->
-	declare_definition_real x m
-    | PAp(PAp(PEq,PName(x)),m) ->
-	declare_definition_real x m
-    | _ -> raise (GenericError ("Incorrect format for definition " ^ name))
-  end
 
-let declare_thf_logic_formula (name:string) (role:string) (m:pretrm) =
+let rec declare_thf_logic_formula (name:string) (role:string) (m:pretrm) =
   begin
     if !verbosity > 4 then Printf.printf "declare_thf_logic_formula %s %s\n" name role;
     if ((role = "axiom") || (role = "hypothesis") || (role = "lemma")) then
@@ -756,12 +801,30 @@ let declare_thf_logic_formula (name:string) (role:string) (m:pretrm) =
 	      let ntm = Ap(Neg,tm) in
 	      let ntmn = norm name_def (logicnorm ntm) in
 	      initial_branch_prenorm := (ntm::(!initial_branch_prenorm));
-	      initial_branch := (ntmn::(!initial_branch))	      
+	      initial_branch := (ntmn::(!initial_branch));
+	      Hashtbl.add part_of_conjecture ntmn ()
 	    end
       end
     else
       raise (GenericError ("Unknown role " ^ role))
   end
+and declare_definition (name:string) (role:string) (m:pretrm) =
+  try
+    begin
+      if (get_bool_flag "ALL_DEFS_AS_EQNS") then raise TreatAsAssumption;
+      if !verbosity > 4 then Printf.printf "declare_definition %s %s\n" name role;
+      match m with
+	PDef(PName(x),m) -> (*** No longer THF syntax. ***)
+	  declare_definition_real x m
+      | PAp(PAp(PEq,PName(x)),m) ->
+	  declare_definition_real x m
+      | _ -> (*** Treat as an assumption, now matter how it looks. Odd, but OK. This may be too liberal; we haven't decided yet. ***)
+	  raise TreatAsAssumption
+(*** raise (GenericError ("Incorrect format for definition " ^ name)) ***)
+    end
+  with
+  | TreatAsAssumption ->
+      declare_thf_logic_formula name "hypothesis" m
 
 (*** Code for enumeration of types and terms - Dec 10, 2010 - Chad ***)
 let enum_started = ref false
@@ -817,10 +880,12 @@ let search_init () =
 (*** reset search ***)
 let reset_search () =
   begin
+    Hashtbl.clear clausesTable;
     clauses := [];
     clause_ruleinfo := Hashtbl.create 100;
     PQueue0.reset();
     PQueue1.reset();
+    PQueue2.reset();
     Hashtbl.clear patoms;
     Hashtbl.clear natoms;
     Hashtbl.clear pchoiceatoms;
