@@ -10,8 +10,8 @@ exception Redeclaration of string
 exception FileNotFound of string
 exception GenericError of string
 
-val mode : string option ref
-val timeout : int option ref
+val mode : string list ref
+val timeout : float option ref
 val verbosity : int ref
 val coq : bool ref
 val coqchannel : out_channel option ref
@@ -27,7 +27,7 @@ val coqsig_def_trm : (string * trm) list ref
 val coqsig_hyp_trm : (string * trm) list ref
 val name_base : (string,unit) Hashtbl.t
 val name_tp : (string,stp) Hashtbl.t
-val name_trm : (string,trm * stp) Hashtbl.t
+val name_trm : (string,(trm * stp) * bool ref) Hashtbl.t
 val name_def : (string,trm) Hashtbl.t
 val name_def_prenorm : (string,trm) Hashtbl.t
 val name_hyp : (string,trm) Hashtbl.t
@@ -112,6 +112,8 @@ val print_refut : refut -> unit (*** Only for debugging. ***)
 val processed : (trm,unit) Hashtbl.t
 val clause_ruleinfo : (clause,ruleinfo) Hashtbl.t ref
 
+exception DuplicateClause
+
 val insert_assumption_clause : clause -> unit
 val insert_search_clause : clause -> ruleinfo option -> unit
 
@@ -136,6 +138,8 @@ val univpreds : (stp,(int * trm)) Hashtbl.t
 
 val filtered : (int,unit) Hashtbl.t
 
+val part_of_conjecture : (trm,unit) Hashtbl.t
+
 val set_default_elt : string -> trm -> unit
 val default_elt : string -> trm
 val default_elt_p : string -> bool
@@ -144,11 +148,14 @@ val get_instantiations : stp -> trm list
 val known_instantiation : stp -> trm -> bool
 val add_instantiation : stp -> trm -> unit
 
+(*** Hash table associating names of epsilon operators with (a,m) where a is the type and m is the formula giving the choice axiom ***)
+val choiceopnames : (string,(stp * trm)) Hashtbl.t
+
 (*** Check if a formula says that a name is a choice operator at some type ***)
 val choiceop_axiom : trm -> (string * stp) option
 
 (*** Declare a name to be a choice operator at some type ***)
-val declare_choiceop : string -> stp -> unit
+val declare_choiceop : string -> stp -> trm -> unit
 
 (*** If trm is a choice operator at some type, return the type, otherwise None ***)
 val choiceop : trm -> stp option
@@ -170,7 +177,7 @@ val decomposable : string -> bool
 exception Unsatisfiable of refut option
 exception Satisfiable
 
-val get_timeout_default : int -> int
+val get_timeout_default : float -> float
 
 val st_include_fun : (string -> unit) ref
 
