@@ -193,7 +193,7 @@ let coqify_name_1 x =
 	  end
 	else
 	  begin
-	    if (!fst) then r := "_";
+	    if ((!fst) && (co <> 95)) then r := "_";
 	    fst := false;
 	    if ((co >= 48) && (co <= 57)) then
 	      r := (!r) ^ (Char.escaped c)
@@ -314,33 +314,33 @@ let rec print_pretrm_coq c m h hu lp rp =
 	  Printf.fprintf c ")";
 	end
   | PAp(PAp(PAnd,m1),m2) ->
-      if ((lp < 20) && (rp < 21)) then
+      if ((lp < 21) && (rp < 20)) then
 	begin
-	  print_pretrm_coq c m1 h hu lp 20;
+	  print_pretrm_coq c m1 h hu lp 21;
 	  Printf.fprintf c " /\\ ";
-	  print_pretrm_coq c m2 h hu 21 rp;
+	  print_pretrm_coq c m2 h hu 20 rp;
 	end
       else
 	begin
 	  Printf.fprintf c "(";
-	  print_pretrm_coq c m1 h hu (-1) 20;
+	  print_pretrm_coq c m1 h hu (-1) 21;
 	  Printf.fprintf c " /\\ ";
-	  print_pretrm_coq c m2 h hu 21 (-1);
+	  print_pretrm_coq c m2 h hu 20 (-1);
 	  Printf.fprintf c ")";
 	end
   | PAp(PAp(POr,m1),m2) ->
-      if ((lp < 18) && (rp < 19)) then
+      if ((lp < 19) && (rp < 18)) then
 	begin
-	  print_pretrm_coq c m1 h hu lp 18;
+	  print_pretrm_coq c m1 h hu lp 19;
 	  Printf.fprintf c " \\/ ";
-	  print_pretrm_coq c m2 h hu 19 rp;
+	  print_pretrm_coq c m2 h hu 18 rp;
 	end
       else
 	begin
 	  Printf.fprintf c "(";
-	  print_pretrm_coq c m1 h hu (-1) 18;
+	  print_pretrm_coq c m1 h hu (-1) 19;
 	  Printf.fprintf c " \\/ ";
-	  print_pretrm_coq c m2 h hu 19 (-1);
+	  print_pretrm_coq c m2 h hu 18 (-1);
 	  Printf.fprintf c ")";
 	end
   | PAp(PAp(PIff,m1),m2) ->
@@ -775,7 +775,7 @@ let expected_tp m a b =
       if (aa = b) then () else raise (ExpectedTypeError(m,aa,b))
   | _ -> ()
 
-let rec to_trm (h:(string,trm * stp) Hashtbl.t) (g:ctx) (m:pretrm) (a:stp option) =
+let rec to_trm (h:(string,(trm * stp) * bool ref) Hashtbl.t) (g:ctx) (m:pretrm) (a:stp option) =
   match m with
   | PName x ->
       begin(*** look in g, then in h, then fail ***)
@@ -784,9 +784,10 @@ let rec to_trm (h:(string,trm * stp) Hashtbl.t) (g:ctx) (m:pretrm) (a:stp option
 	with
 	| Not_found ->
 	    try
-	      let zz = Hashtbl.find h x in
+	      let (zz,o) = Hashtbl.find h x in
 	      begin
 		(match zz with (_,b) -> expected_tp m a b);
+		o := true; (** Mark that it's occurred. - Chad, March 31, 2011 **)
 		zz
 	      end
 	    with
