@@ -82,7 +82,7 @@ val onlybetanorm:trm -> trm
 val logicnorm:trm -> trm
 
 val to_stp:pretrm -> stp
-val to_trm:(string,trm * stp) Hashtbl.t -> ctx -> pretrm -> stp option -> trm * stp
+val to_trm:(string,(trm * stp) * bool ref) Hashtbl.t -> ctx -> pretrm -> stp option -> trm * stp
 
 val neg_p:trm -> bool
 val neg_body:trm -> trm option
